# coding: utf-8
from __future__ import absolute_import

from functools import partial
from functools import reduce
from functools import total_ordering
from operator import gt
import abc
import itertools
import operator

from django.db import models
from django.db.models import Q
from django.db.models.fields.related import RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.manager import Manager
from django.forms.fields import BooleanField
from django.utils.encoding import force_text
from m3.actions.exceptions import ApplicationLogicException
from six.moves import map
from six.moves import range
from xlsxwriter import Workbook
import six

from educommon.report.constructor.constants import FALSE
from educommon.report.constructor.constants import TRUE
from educommon.report.constructor.utils import get_field
from educommon.report.constructor.utils import get_field_value_by_display
from educommon.utils.misc import cached_property

from ... import constants
from ...base import ColumnDescriptor
from ...constants import DIRECTION_ASC
from ...exceptions import DataSourceParamsNotFound
from ...exceptions import FilterError
from ...exceptions import ReportConstructorException
from ...models import ReportFilter
from ...models import ReportFilterGroup
from ...models import ReportTemplate
from ...registries import registry
from ...utils import get_columns_hierarchy


class FilterValuesProvider(object):

    u"""Поставщик данных фильтра.

    Используется как для формирования Q-фильтра для ORM, так и для
        последующей фильтрации полученных данных.

    Возвращает значение из фильтра, преобразованное к объекту Python.

    Тип возвращаемого значения определяется оператором сравнения, который
    указан в фильтре:

        * **Пусто**: возвращается значения ``None``.
        * **Равно одному из**: возвращает кортеж из элементов с типом,
          соответствующим типу поля, по которому выполняется фильтрация данных.
        * **Между**: возвращает кортеж из двух элементов с типом,
          соответствующим типу поля, по которому выполняется фильтрация данных.
        * Во всех остальных случаях возвращает значение фильтра,
          преобразованное к типу данных того столбца, по которому выполняется
          фильтрация.

    .. note::

       Под значением фильтра подразумевается значение поля ``values`` модели
       :class:`~educommon.report.constructor.models.ReportFilter`.

    """

    def __init__(self, as_orm_filter=True):
        u"""Инициализация провайдера.

        :param bool as_orm_filter: преобразовать значения к типам ORM
            или к типам отчёта (например, True/False или Да/Нет).
        """

        self._as_orm_filter = as_orm_filter

    def to_python(self, value, report_filter, column_descriptor):
        u"""Возвращает значение из фильтра, приведенное к объекту Python.

        В параметрах фильтра значение хранится в виде текстовой строки.
        """

        if column_descriptor.data_type == constants.CT_CHOICES:
            return get_field_value_by_display(column_descriptor.field, value)
        else:
            if isinstance(column_descriptor.field.formfield(), BooleanField):
                if (
                    isinstance(value, six.string_types) and
                    value.lower() in (u'false', u'0', u'нет')
                ):
                    return False if self._as_orm_filter else FALSE

                elif (
                    isinstance(value, six.string_types) and
                    value.lower() in (u'true', u'1', u'да')
                ):
                    return True if self._as_orm_filter else TRUE

                else:
                    raise FilterError(
                        report_filter, u'неправильно задано значение фильтра'
                    )

            return column_descriptor.field.formfield().to_python(value)

    def __call__(self, report_filter, column_descriptor):
        u"""Возвращает значение из фильтра, преобразованное к объекту Python.

        :param report_filter: Фильтр шаблона отчетов.
        :type report_filter: educommon.report.constructor.models.ReportFilter

        :param column_descriptor: Дескриптор поля модели, входящего в состав
            столбца отчета.
        :type column_descriptor:
            educommon.report.constructor.base.ColumnDescriptor
        """
        if report_filter.operator == constants.IS_NULL:
            result = None

        elif report_filter.operator == constants.IN:
            result = tuple(
                self.to_python(value, report_filter, column_descriptor)
                for value in report_filter.values
            )

        elif report_filter.operator == constants.BETWEEN:
            result = tuple(
                self.to_python(value, report_filter, column_descriptor)
                for value in report_filter.values
            )

            if len(result) != 2:
                raise FilterError(
                    report_filter, u'не правильно заданы границы диапазона'
                )

        elif report_filter.values:
            result = self.to_python(
                report_filter.values[0],
                report_filter,
                column_descriptor
            )

        else:
            raise FilterError(
                report_filter, u'не указано значение для сравнения.'
            )

        return result


def is_row(data):
    u"""Возвращает True, если в ``data`` содержится строка с данными.

    :rtype: bool
    """
    return (
        isinstance(data, (tuple, list)) and
        any(not isinstance(cell, (tuple, list)) for cell in data)
    )


def is_block(data):
    u"""Возвращает True, если в ``data`` содержится блок записей.

    :rtype: bool
    """
    return (
        data and
        isinstance(data, (tuple, list)) and
        all(isinstance(row, (tuple, list)) for row in data)
    )


def get_data_width(data):
    u"""Возвращает количество колонок, занимаемое данными.

    :param list data: Данные блока, строки или ячейки.
        - Ячейка --- это простые данные (строка, число, дата и т.п.).
        - Строка --- это кортеж из данных ячеек и вложенных блоков.
        - Блок --- это кортеж из строк. Содержит только строки.

    :rtype: int
    """
    if is_block(data):
        return max(map(get_data_width, data))
    elif is_row(data):
        return sum(map(get_data_width, data))
    else:
        return 1


class _FilterBuilder(object):

    u"""Построитель фильтров для выборки данных из БД."""

    def __init__(self, report_filter, column_descriptor):
        u"""Инициализация экземпляра.

        :param report_filter: Фильтр шаблона отчетов.
        :type report_filter: educommon.report.constructor.models.ReportFilter

        :param column_descriptor: Дескриптор поля модели, входящего в состав
            столбца отчета.
        ::type column_descriptor:
            educommon.report.constructor.base.ColumnDescriptor
        """
        assert isinstance(report_filter, ReportFilter), type(report_filter)
        assert isinstance(column_descriptor, ColumnDescriptor), type(
            column_descriptor
        )

        self._report_filter = report_filter
        self._column_descriptor = column_descriptor

    def _get_lookup(self, field_lookup):
        u"""Возвращает lookup-выражение, соответствующее параметрам фильтра.

        При формировании lookup-выражения учитываются параметры фильтра
        ``case_sensitive``, ``operator``, а также тип данных столбца, по
        которому выполняется фильтрация.

        :param str field_lookup: lookup-выражение для доступа к значению поля
        """
        if (
            not self._report_filter.case_sensitive and
            self._column_descriptor.data_type == constants.CT_TEXT
        ):
            field_lookup += {
                constants.EQ: '__iexact',
                constants.CONTAINS: '__icontains',
                constants.STARTS_WITH: '__istartswith',
                constants.ENDS_WITH: '__iendswith',
                constants.IN: '__lower__in',
            }.get(self._report_filter.operator, '')

        else:
            field_lookup += {
                constants.LE: '__lte',
                constants.LT: '__lt',
                constants.GT: '__gt',
                constants.GE: '__gte',
                constants.CONTAINS: '__contains',
                constants.STARTS_WITH: '__startswith',
                constants.ENDS_WITH: '__endswith',
                constants.BETWEEN: '__range',
                constants.IN: '__in',
            }.get(self._report_filter.operator, '')

        return field_lookup

    def _get_values(self):
        if self._report_filter.operator == constants.IS_NULL:
            values = True

        elif (
            not self._report_filter.case_sensitive and
            self._column_descriptor.data_type == constants.CT_TEXT and
            self._report_filter.operator == constants.IN
        ):
            values = tuple(v.lower() for v in self._report_filter.values)

        else:
            values = FilterValuesProvider()(
                self._report_filter, self._column_descriptor
            )

        return values

    def get_orm_filter(self):
        u"""Возвращает условия выборки из шаблона отчета.

        :rtype: django.db.models.Q
        """
        field_lookup = self._column_descriptor.full_lookup

        filter_values = self._get_values()

        if self._report_filter.operator == constants.IS_NULL:
            if self._column_descriptor.data_type == constants.CT_TEXT:
                result = Q(
                    Q(**{field_lookup + '__isnull': True}) |
                    Q(**{field_lookup: u''})
                )
            else:
                result = Q(**{field_lookup + '__isnull': True})

        else:
            try:
                result = Q(**{
                    self._get_lookup(field_lookup): filter_values,
                })
            except ValueError as error:
                raise FilterError(self._report_filter, six.text_type(error))

        if self._report_filter.exclude:
            result = ~result

        return result


class _FilterGroupBuilder(object):

    u"""Построитель Q-выражений и фильтрующих функций для групп фильтров."""

    def __init__(self, data_filterer, filter_group):
        u"""Инициализация экземпляра класса.

        :param data_filterer: Фильтратор данных отчета.
        :type data_filterer: _DataFilterer

        :param filter_group: Группа фильтров.
        :type filter_group:
            educommon.report.constructor.models.ReportFilterGroup
        """
        assert isinstance(data_filterer, _DataFilterer), type(data_filterer)
        assert isinstance(filter_group, ReportFilterGroup), type(
            filter_group
        )

        self._data_filterer = data_filterer
        self._filter_group = filter_group

    def _get_nested_orm_filters(self):
        for nested_group in self._filter_group.nested_groups.all():
            if any(
                report_filter
                for report_filter in six.itervalues(self._data_filterer)
                if report_filter.group_id == nested_group.id
            ):
                yield nested_group.get_orm_filter()

        data_source = self._data_filterer.data_source_descriptor

        for report_filter in six.itervalues(self._data_filterer.filters_by_id):
            if report_filter.group_id == self._filter_group.id:
                column_descriptor = data_source.get_column_descriptor(
                    report_filter.column.name
                )
                if isinstance(column_descriptor.field, models.Field):
                    filter_builder = _FilterBuilder(
                        report_filter, column_descriptor
                    )
                    yield filter_builder.get_orm_filter()

    def get_orm_filter(self):
        u"""Возвращает Q-выражение для данной группы фильтров.

        :rtype: django.db.models.Q
        """
        operators = {
            ReportFilterGroup.OPERATOR_AND: operator.and_,
            ReportFilterGroup.OPERATOR_OR: operator.or_,
        }

        if self._filter_group.operator not in operators:
            raise ApplicationLogicException(
                u'Неподдерживаемый оператор: {}'
                .format(self._filter_group.operator)
            )

        orm_filters = tuple(self._get_nested_orm_filters())

        if orm_filters:
            result = reduce(
                operators[self._filter_group.operator],
                orm_filters
            )
        else:
            result = Q()

        return result


class _DataFilterer(object):

    u"""Фильтратор данных отчета.

    Фильтрация данных выполняется в два этапа:

        1. Наложение фильтров на запрос.
        2. Фильтрация полученных данных в приложении.
    """

    def __init__(self, report_template, data_source_descriptor,
                 report_columns, ignored_columns_ids):
        u"""Инициализация экземпляра класса.

        :param report_template: Шаблон фильтра.
        :type report_template:
            educommon.report.constructor.models.ReportTemplate

        :param data_source_descriptor: Дескриптор источника данных..
        :type data_source_descriptor:
            educommon.report.constructor.base.ModelDataSourceDescriptor

        :param report_columns: Колонки отчета, упорядоченные по порядковому
            номеру.
        :type: tuple
        :param ignored_columns_ids: Колонки исключенные из отчета.
        :type: tuple
        """
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        self._report_template = report_template
        self.data_source_descriptor = data_source_descriptor
        self._report_columns = report_columns
        self._ignored_columns_ids = ignored_columns_ids

    @cached_property
    def filters_by_id(self):
        u"""Все фильтры шаблона по id фильтра.

        :rtype: dict
        """
        query = ReportFilter.objects.filter(
            group__report_template=self._report_template,
        ).exclude(
            column_id__in=self._ignored_columns_ids
        )

        return {
            report_filter.id: report_filter
            for report_filter in query
        }

    def get_orm_filters(self):
        u"""Возвращает фильтры для формирования SQL-запроса.

        .. note::

           Предназначена для предварительной фильтрации данных на уровне СУБД.
           Т.к. данные загружаются только для модели источника данных без
           данных зависимых объектов, эти фильтры не срабатывают при загрузке
           зависимых объектов. Поэтому дополнительно нужно применять функцию
           фильтрации, возвращаемую методом :meth:`get_filter_function`.

        :rtype: django.db.models.Q
        """
        filter_group = self._report_template.filter_groups.filter(
            parent__isnull=True
        ).first()

        if filter_group:
            return _FilterGroupBuilder(self, filter_group).get_orm_filter()
        else:
            return Q()

    def _get_function_for_filter(self, report_filter):
        u"""Возвращает функцию для фильтрации данных в соответствии с фильтром.

        :rtype: callable
        """
        data_source = self.data_source_descriptor
        column_descriptor = data_source.get_column_descriptor(
            report_filter.column.name
        )
        filter_values = FilterValuesProvider(as_orm_filter=False)(
            report_filter, column_descriptor
        )

        if report_filter.operator == constants.LE:
            def function(value):
                return value is not None and value <= filter_values

        elif report_filter.operator == constants.LT:
            def function(value):
                return value is not None and value < filter_values

        elif report_filter.operator == constants.EQ:
            def function(value):
                return value == filter_values

        elif report_filter.operator == constants.GT:
            def function(value):
                return value is not None and value > filter_values

        elif report_filter.operator == constants.GE:
            def function(value):
                return value is not None and value >= filter_values

        elif report_filter.operator == constants.IS_NULL:
            def function(value):
                return value is None or value == ''

        elif report_filter.operator == constants.CONTAINS:
            assert isinstance(filter_values, six.string_types), type(
                filter_values
            )

            if report_filter.case_sensitive:
                def function(value):
                    return filter_values in value
            else:
                def function(value):
                    return any(
                        v.lower() in value
                        for v in filter_values
                    )

        elif report_filter.operator == constants.STARTS_WITH:
            assert isinstance(filter_values, six.string_types), type(
                filter_values
            )

            if report_filter.case_sensitive:
                def function(value):
                    return value.startswith(filter_values)
            else:
                def function(value):
                    return value.lower().startswith(filter_values.lower())

        elif report_filter.operator == constants.ENDS_WITH:
            assert isinstance(filter_values, six.string_types), type(
                filter_values
            )

            if report_filter.case_sensitive:
                def function(value):
                    return value.endswith(filter_values)
            else:
                def function(value):
                    return value.lower().endswith(filter_values.lower())

        elif report_filter.operator == constants.BETWEEN:
            if report_filter.case_sensitive:
                def function(value):
                    return (
                        value is not None and
                        filter_values[0] <= value <= filter_values[1]
                    )
            else:
                def function(value):
                    if isinstance(value, six.text_type):
                        value = value.lower()
                        values = tuple(v.lower() for v in filter_values)
                    else:
                        values = filter_values

                    return (
                        value is not None and
                        values[0] <= value <= values[1]
                    )

        elif report_filter.operator == constants.IN:
            assert isinstance(filter_values, tuple), type(filter_values)

            if report_filter.case_sensitive:
                def function(value):
                    return value in filter_values
            else:
                def function(value):
                    return value.lower() in (v.lower() for v in filter_values)

        else:
            raise FilterError(
                report_filter,
                u'Неподдерживаемый оператор ({})'.format(
                    report_filter.operator
                )
            )

        if report_filter.exclude:
            return lambda value: not function(value)
        else:
            return function

    def get_filter_function(self):
        u"""Возвращает функцию для фильтрации данных.

        Эта функция для каждого набора данных возвращает новый набор данных,
        которые отфильтрованы в соответствии с параметрами фильтрации,
        указанными в шаблоне отчета.

        Запись может быть отфильтрована полностью, либо отфильтрованы только
        вложенные блоки. В первом случае функция возвращает ``None``. Во
        втором случае --- будут удалены записи во внутренних блоках.
        """
        filter_functions = {
            report_filter.column_id: self._get_function_for_filter(
                report_filter
            )
            for report_filter in six.itervalues(self.filters_by_id)
        }
        filter_functions_by_column = tuple(
            filter_functions.get(report_column.pk)
            for report_column in self._report_columns
        )

        def filter_function(row_data, column_functions=None):
            u"""Возвращает отфильтрованные данные записи отчета."""
            if column_functions is None:
                column_functions = filter_functions_by_column

            result = []
            for cell in row_data:
                cell_width = get_data_width(cell)
                column_filter_functions = column_functions[:cell_width]
                column_functions = column_functions[cell_width:]

                if is_block(cell):
                    filtered_block = _filter_block_or_row(
                        cell, column_filter_functions)
                    if filtered_block is None:
                        return None
                    result.append(filtered_block)
                else:
                    # Простое значение (не строка или блок)
                    assert not is_row(cell), cell
                    column_filter_function = column_filter_functions[0]
                    if (
                            isinstance(cell, six.string_types) and
                            column_filter_function and
                            not column_filter_function(cell)
                    ):
                        return None
                    result.append(cell)

            return result

        def _filter_block_or_row(block_or_row, column_filter_functions):
            u"""Возвращет отфильтрованный блок/строку или None."""
            if is_row(block_or_row):
                return filter_function(block_or_row, column_filter_functions)
            filtered = [_filter_block_or_row(item, column_filter_functions)
                        for item in block_or_row]
            filtered = [item for item in filtered if item is not None]
            return filtered or None

        return filter_function


@total_ordering
class _OrderInverter(object):

    """Класс-обёртка для смены направления сортировки."""

    # pylint: disable=eq-without-hash

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        return gt(self.value, other.value)


class _DataSorter(object):

    """Сортировщик данных отчёта.

    Структура данных для отчёта представляет из себя набор кортежей (см.
    описание класса ``DataLoader``). Все кортежи можно разделить на два типа:
    строка и блок. Блок --- это набор строк, строка, в свою очередь, может
    содержать как данные ячеек, так и другие блоки. Если строка содержит блоки,
    это означает, что объекты-источники, содержащие соответствующие данные,
    связаны между собой отношением один-ко-многим.

    Сортировка строк осуществляется внутри блока на основе данных в ячейках, но
    не на основе данных вложенных блоков. Это означает, что если в параметрах
    сортировки указаны столбцы, попадающие на вложенные блоки, то сортировка
    будет внутри этих блоков, но на порядок строк во внешнем блоке это не
    повлияет.
    """

    def __init__(self, report_template, ignored_columns_ids):
        """Инициализация сортировщика данных.

        :param report_template: шаблон отчета.
        :type report_template:
            educommon.report.constructor.models.ReportTemplate

        :param ignored_columns_ids: идентификаторы игнорируемых столбцов (эти
            столбцы были добавлены в шаблон отчета до того, как были удалены
            из моделей Системы).
        """
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        self._report_template = report_template
        self._ignored_columns_ids = ignored_columns_ids

    @cached_property
    def _params(self):
        data = enumerate(
            self._report_template.columns.exclude(
                id__in=self._ignored_columns_ids
            ).order_by(
                'index',
            ).values_list(
                'sorting__index',
                'sorting__direction',
            )
        )

        result = {
            index: {
                'sort_index': sort_index,
                'sort_direction': direction,
            }
            for index, (sort_index, direction) in data
            if sort_index is not None
        }

        return result

    def _get_sort_key(self, row, start_index):
        """Возвращает ключ для сортировки строки."""
        assert isinstance(row, list), (type(row), row)

        values = {}
        index = start_index

        for cell in row:
            if is_block(cell):
                index += get_data_width(cell)

            else:
                if index in self._params:
                    # Для этого столбца определены параметры сортировки.
                    sort_index = self._params[index]['sort_index']
                    direction = self._params[index]['sort_direction']
                    if direction == DIRECTION_ASC:
                        values[sort_index] = cell
                    else:
                        values[sort_index] = _OrderInverter(cell)
                index += 1

        result = tuple(
            values[sort_index]
            for sort_index in sorted(values)
        )

        return result

    def sort(self, rows, start_index=0):
        """Возвращает отсортированные данные отчета.

        :param rows: данные отчета (описание структуры см. в документации
            класса ``DataLoader``).

        :rtype: list
        """
        for row in rows:
            assert isinstance(row, list), (type(row), row)
            index = start_index
            for cell_index, cell in enumerate(row):
                if is_block(cell):
                    row[cell_index] = self.sort(cell, index)
                    index += get_data_width(cell)
                else:
                    index += 1

        rows = sorted(
            rows,
            key=partial(self._get_sort_key, start_index=start_index),
        )

        return rows


class DataLoader(object):

    u"""Загрузчик данных для шаблона отчета.

    Данные каждой строки отчета представлены в виде кортежа. Элементами кортежа
    могут быть значения простых типов (числа, строки и т.п.), а также
    кортежи кортежей (вложенные блоки). Вложенные блоки добавляются в тех
    случаях, когда в параметрах отчета указаны т.н. обратные связи (связь "один
    ко многим"), когда одной записи в модели источника данных соответствует
    несколько записей в связанных моделях. В таких вложенных кортежах также
    могут быть простые данные и вложенные кортежи кортежей. Глубина вложенности
    не ограничивается и зависит от настроек шаблона отчета.

    Пример:

    .. table:: Данные в табличном виде
       :align: center

       +-------------+---------+---------------------------------------------+
       | **Фамилия** | **Имя** |                **Учащиеся**                 |
       |             |         +---------------------------------------------+
       |             |         |                 **Группа**                  |
       |             |         +--------------------------+------------------+
       |             |         |       **Учреждение**     | **Наименование** |
       |             |         +--------------------------+                  |
       |             |         | **Краткое наименование** |                  |
       +-------------+---------+--------------------------+------------------+
       | Иванов      | Иван    | СДЮШОР1                  | Бокс             |
       |             |         +--------------------------+------------------+
       |             |         | СДЮШОР1                  | Фехтование       |
       +-------------+---------+--------------------------+------------------+

    Одна запись в приведенной выше таблице будет иметь следующую структуру:

    .. code-block:: python

       (
           'Иванов',
           'Иван',
           (
               ('СДЮШОР1', 'Бокс'),
               ('СДЮШОР1', 'Фехтование'),
           )
       )
    """

    def __init__(self, report_template, data_source, report_columns,
                 ignored_columns_ids, user):
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )
        self._report_template = report_template
        self._data_source = data_source
        self._report_columns = report_columns
        self._ignored_columns_ids = ignored_columns_ids
        self._user = user

    @staticmethod
    def _get_column_count(columns):
        # Для иерархии столбцов возвращает количество занимаемых столбцов.
        if columns:
            return sum(map(
                DataLoader._get_column_count, six.itervalues(columns)
            ))
        else:
            return 1

    @staticmethod
    def _get_field_display(field, field_value):
        for value, name in field.flatchoices:
            if value == field_value:
                return force_text(name, strings_only=True)

    @staticmethod
    def _get_object_data(obj, attr_name, nested):
        field = None
        if attr_name:
            field = get_field(obj, attr_name)
            model = field.model
            if isinstance(field, ForeignObjectRel):
                descriptor = getattr(model, field.get_accessor_name())
            elif isinstance(field, RelatedField):
                descriptor = getattr(model, field.name)
            else:
                descriptor = None

            if descriptor:
                try:
                    attr_value = getattr(obj, attr_name)
                except descriptor.RelatedObjectDoesNotExist:
                    attr_value = None
            else:
                try:
                    attr_value = getattr(obj, attr_name)
                except AttributeError:
                    if (
                        hasattr(obj, 'report_constructor_params') and
                        'extra' in obj.report_constructor_params and
                        attr_name in obj.report_constructor_params['extra']
                    ):
                        # Дополнительное поле.
                        extra = obj.report_constructor_params['extra']
                        params = extra[attr_name]
                        if 'attr' in params:
                            attr_value = getattr(obj, params['attr'])
                            if callable(attr_value):
                                attr_value = attr_value()
                        elif 'func' in params:
                            attr_value = params['func'](obj)
                        else:
                            raise

                if (
                    attr_value is not None and
                    isinstance(field, (models.BooleanField,
                                       models.NullBooleanField))
                ):
                    attr_value = (
                        TRUE if attr_value else FALSE
                    )
        else:
            attr_value = obj
            descriptor = None

        # Обратная связь или M2M. Такие данные выводятся в виде кортежа.
        if (
            attr_name and
            (
                isinstance(field, ForeignObjectRel) and
                hasattr(descriptor, 'related_manager_cls') and
                isinstance(attr_value, descriptor.related_manager_cls) or
                isinstance(field, models.ManyToManyField) and
                isinstance(attr_value, Manager)
            )
        ):
            objects = attr_value.all()
            if objects:
                related_obj_data_gen = (
                    list(DataLoader._get_object_data(related_obj, None, nested))
                    for related_obj in objects
                )
                yield list(itertools.takewhile(bool, related_obj_data_gen))
            else:
                yield [[None] * DataLoader._get_column_count(nested)]

        # Прямая связь (ForeignKey или OneToOneField).
        elif nested:
            if attr_value:
                for k, v in six.iteritems(nested):
                    for object_data in (
                        DataLoader._get_object_data(attr_value, k, v)
                    ):
                        yield object_data
            else:
                for _ in range(DataLoader._get_column_count(nested)):
                    yield None

        else:
            # Поле, значения которого определяются через внутренный base_field
            if getattr(field, 'base_field', False) and attr_value:
                inner_values = []
                for inner_value in attr_value:
                    if getattr(field.base_field, 'flatchoices', False):
                        inner_values.append(
                            DataLoader._get_field_display(
                                field.base_field, inner_value
                            )
                        )
                    else:
                        inner_values.append(inner_value)
                yield list(
                    six.text_type(value) for value in inner_values
                )

            # Поле с choices.
            elif getattr(field, 'flatchoices', False):
                yield DataLoader._get_field_display(
                    field, attr_value
                )

            # Поле с данными.
            else:
                yield attr_value

    @cached_property
    def _data_filterer(self):
        u"""Фильтратор данных отчета."""
        return _DataFilterer(
            self._report_template, self._data_source, self._report_columns,
            self._ignored_columns_ids
        )

    def _available_units_filter(self, query):
        u"""Фильтратор данных отчета по доступности учреждений."""
        return self._data_source.add_source_filter(
            query, self._report_template.include_available_units,
            self._user
        )

    def _get_objects(self):
        result = self._data_source.model.objects.all()
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Добавление фильтров в запрос.

        if ReportFilter.objects.filter(
            column__report_template=self._report_template,
        ).exclude(
            column_id__in=self._ignored_columns_ids
        ).exists():
            result = result.filter(
                self._data_filterer.get_orm_filters()
            )
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Добавление параметров включения вложенных учреждений.

        result = self._available_units_filter(result)
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        return result

    def _get_rows_data(self, columns_hierarchy):
        filter_function = self._data_filterer.get_filter_function()

        for obj in self._get_objects():
            row_data = list(
                DataLoader._get_object_data(obj, None, columns_hierarchy)
            )
            if row_data and any(row_data):
                row_data = filter_function(row_data)
                if row_data:
                    yield row_data

    def __iter__(self):
        column_names = self._report_template.columns.filter(
            visible=True,
        ).exclude(
            pk__in=self._ignored_columns_ids
        ).values_list('name', flat=True)

        if not column_names:
            raise ReportConstructorException(
                u'В шаблоне нет ни одного отображаемого столбца.'
            )

        columns_hierarchy = get_columns_hierarchy(*column_names)

        result = list(self._get_rows_data(columns_hierarchy))

        sorter = _DataSorter(self._report_template, self._ignored_columns_ids)
        result = sorter.sort(result)

        return iter(result)


class ReportBuilderBase(six.with_metaclass(abc.ABCMeta, object)):
    u"""Базовый класс построителя отчета."""

    def __init__(self, report_template, file_path, user):
        u"""Инициализация класса.

        :param report_template: Шаблон отчета.
        :type report_template:
            educommon.report.constructor.models.ReportTemplate

        :param unicode file_path: Путь к файлу отчета.
        """
        assert isinstance(report_template, ReportTemplate), type(
            report_template
        )

        if report_template.data_source_name not in registry:
            raise DataSourceParamsNotFound(report_template.data_source_name)

        self._report_template = report_template
        self._file_path = file_path
        self._user = user

    @cached_property
    def _data_source(self):
        u"""Имя источника данных, указанного в шаблоне.

        :rtype: unicode
        """
        return self._report_template.data_source

    @cached_property
    def _report_columns(self):
        u"""Столбцы отчета.

        :rtype: tuple
        """
        return tuple(
            self._report_template.columns.exclude(
                pk__in=self._ignored_columns_ids
            )
        )

    @cached_property
    def _ignored_columns_ids(self):
        u"""Исключенные из отчета колонки

        :rtype: set
        """
        ignored_columns = set()
        for col_id, col_name in self._report_template.columns.values_list(
                'id', 'name'):
            if self._data_source.is_column_ignored(col_name):
                ignored_columns.add(col_id)

        return ignored_columns

    @cached_property
    def _data(self):
        u"""Отфильтрованные и отсортированные данные отчета.

        :rtype: list
        """
        return list(
            DataLoader(
                self._report_template, self._data_source, self._report_columns,
                self._ignored_columns_ids, self._user
            )
        )

    @cached_property
    def _workbook(self):
        u"""Книга Excel, в которой формируется отчет."""
        result = Workbook(self._file_path, dict(
            default_date_format='dd.mm.yyyy',
        ))

        for cell_format in (result.default_date_format,
                            result.default_url_format):
            cell_format.set_border()
            cell_format.set_align('vcenter')

        return result

    @cached_property
    def _worksheet(self):
        u"""Страница в книге Excel."""
        return self._workbook.add_worksheet(u'Отчет')

    @abc.abstractmethod
    def _flush_header(self):
        u"""Запись заголовка отчета."""

    @abc.abstractmethod
    def _flush_data(self):
        u"""Запись данных в отчет."""

    def build(self):
        u"""Формирование отчета."""
        self._flush_header()
        self._flush_data()

        self._workbook.close()
